<?php

declare(strict_types = 1);

namespace Drupal\smart_migrate_cli\Commands;

use Consolidation\SiteAlias\SiteAliasManagerAwareInterface;
use Consolidation\SiteAlias\SiteAliasManagerAwareTrait;
use Consolidation\SiteProcess\SiteProcess;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Database;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\Core\Session\UserSession;
use Drupal\Core\Site\Settings;
use Drupal\Core\State\StateInterface;
use Drupal\migrate\Plugin\MigrationInterface;
use Drupal\smart_migrate_cli\SimpleMigrationManager;
use Drupal\smart_migrate_cli\Traits\SmartMigrateHelperTrait;
use Drupal\smart_migrate_cli\Traits\SmartMigrationConfigurationTrait;
use Drush\Commands\DrushCommands;
use Drush\Exceptions\UserAbortException;
use Drush\Utils\StringUtils;
use Symfony\Component\Console\Command\LockableTrait;
use Symfony\Component\Console\Helper\ProgressBar;
use Symfony\Component\Console\Output\OutputInterface;
use Webmozart\PathUtil\Path;

/**
 * Smart Migrate CLI's own Drush commands.
 */
class SmartMigrateCliCommands extends DrushCommands implements SiteAliasManagerAwareInterface {

  use SiteAliasManagerAwareTrait;
  use SmartMigrateHelperTrait;
  use SmartMigrationConfigurationTrait;
  use LockableTrait;

  /**
   * Identifier of the main multi thread migration process.
   *
   * Used for locking and killing subprocesses.
   *
   * @const string
   */
  const MT_ID = 'mtImport';

  /**
   * Environment variable name storing migration plugin ID.
   *
   * @const string
   */
  const SMC_MIGRATION_PLUGIN_ID_ENV = 'SMC_MIGRATION_PLUGIN_ID';

  /**
   * Exit code used of processes which migration's requirements aren't met.
   *
   * @const int
   */
  const EXIT_REQUIREMENTS_NOT_MET = 10;

  /**
   * Exit code of single processes which migration threw an exception.
   *
   * @const int
   */
  const EXIT_PROCESS_CRITICAL = 100;

  /**
   * Minimum time to wait between updating scheduled processes and progress.
   *
   * @const float
   */
  const MT_POLL_WAIT_SECONDS = 0.2;

  /**
   * Text displayed in empty progress bar sections.
   *
   * @const string
   */
  const EMPTY_PROGRESS_BAR_TEXT = ' [- Idle ---------------------]';

  /**
   * Width of progress bar.
   *
   * @const int
   */
  const PROGRESS_BAR_WIDTH = 28;

  /**
   * ID of the normal progress bar template.
   *
   * @const string
   */
  const MT_PROGRESS_BAR = 'mt_progress';

  /**
   * ID of the "no max" progress bar template.
   *
   * @const string
   */
  const MT_PROGRESS_BAR_NOMAX = 'mt_progress_nomax';

  /**
   * The default progress bar format used by multi thread migrate import.
   *
   * @const string
   */
  const PROGRESS_BAR_FORMAT = ' [%bar%] %percent:3s%% | %current:4s%/%max:-4s% | %message%';

  /**
   * The logger section used by SmartMigrateHelper's getNextUnblockedMigration.
   *
   * Yepp, bad design.
   *
   * @var \Symfony\Component\Console\Output\OutputInterface|null
   */
  private ?OutputInterface $loggerSection;

  /**
   * Storage for empty "placeholder" sections around progress bars.
   *
   * @var \Symfony\Component\Console\Output\OutputInterface[]
   */
  private array $barPlaceholderSections = [];

  /**
   * Error messages recorded during multi-thread migration.
   *
   * First level keys are migration plugin IDs, second level keys are md5 hashes
   * of the messages. Each value is an array where the "message" key contains
   * the message and the "count" key's value is the number of how many times the
   * error was triggered during migration of the specified plugin instance.
   *
   * @var array[][]
   */
  private array $errorOutputs = [];

  /**
   * The progress bars and their metadata.
   *
   * @var array[]
   *
   * Each "subthread" has a progress bar.
   *
   * @todo Provide our own object for these things.
   */
  private array $progressBars = [];

  /**
   * The scheduled subprocesses.
   *
   * @var \Symfony\Component\Process\Process[]
   *
   * @todo Provide our own object for these things too.
   */
  private array $activeProcesses = [];

  /**
   * List of the migration plugin instances to execute.
   *
   * @var \Drupal\migrate\Plugin\MigrationInterface[]
   */
  private array $list = [];

  /**
   * List of the migration plugin IDs which should be executed.
   *
   * @var string[]
   */
  private array $migrationsToProcess = [];

  /**
   * List of the already processed migrations' plugin IDs.
   *
   * @var string[]
   */
  private array $migrationsProcessed = [];

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected StateInterface $state;

  /**
   * Account switcher.
   *
   * @var \Drupal\Core\Session\AccountSwitcherInterface
   */
  protected AccountSwitcherInterface $accountSwitcher;

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Light-weight migration plugin manager service optimized for multi-thread.
   *
   * @var \Drupal\smart_migrate_cli\SimpleMigrationManager|null
   */
  protected SimpleMigrationManager $migrationPluginManager;

  /**
   * Constructs a new SmartMigrateCliCommands instance.
   *
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\Session\AccountSwitcherInterface $account_switcher
   *   The account switcher service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity_type manager.
   * @param \Drupal\smart_migrate_cli\SimpleMigrationManager $simple_migration_plugin_manager
   *   The light-weight migration plugin manager.
   */
  public function __construct(StateInterface $state, AccountSwitcherInterface $account_switcher, EntityTypeManagerInterface $entity_type_manager, SimpleMigrationManager $simple_migration_plugin_manager = NULL) {
    parent::__construct();
    $this->state = $state;
    $this->accountSwitcher = $account_switcher;
    $this->entityTypeManager = $entity_type_manager;
    $this->migrationPluginManager = $simple_migration_plugin_manager;
    $this->loggerSection = $this->output();
  }

  /**
   * Executes migrations in the optimal order, in multiple threads.
   *
   * @param array|null $options
   *   Command line options.
   *
   * @command smart-migrate:multi-thread-import
   *
   * @option tag
   *   Tag of the migrations to execute.
   * @option threads
   *   Maximal number of threads to use. Optional, defaults to 4.
   * @option memory
   *   Display memory usage even if verbosity is normal.
   * @option estimate
   *   Display estimated time even if verbosity is normal.
   * @option run-at-the-end
   *   Execute the specified migrations at the end of the process.
   *
   * @validate-module-enabled migrate
   *
   * @aliases smimt,smmti
   *
   * @throws \Exception
   *   When not enough options were provided or no migration was found.
   */
  public function mtImport(?array $options = [
    'tag' => self::REQ,
    'threads' => self::OPT,
    'memory' => self::OPT,
    'estimate' => self::OPT,
    'run-at-the-end' => self::OPT,
  ]): int {
    if (!$this->lock(self::MT_ID)) {
      $this->logger()->error('The command is already running in another process.');
      return self::EXIT_FAILURE;
    }

    // Initialize our custom progress bar format.
    $format_template = $format_template_nomax = static::PROGRESS_BAR_FORMAT;
    if (
      $this->output()->getVerbosity() > OutputInterface::VERBOSITY_NORMAL ||
      $options['estimate']
    ) {
      $format_template .= ' | %remaining:7s%';
    }
    if (
      $this->output()->getVerbosity() > OutputInterface::VERBOSITY_NORMAL ||
      $options['memory']
    ) {
      $format_template .= ' | %memory:6s%';
      $format_template_nomax .= ' | %memory:6s%';
    }
    ProgressBar::setFormatDefinition(static::MT_PROGRESS_BAR, $format_template);
    ProgressBar::setFormatDefinition(static::MT_PROGRESS_BAR_NOMAX, $format_template_nomax);

    $infosection = $this->output()->section();
    $infosection->writeln('Getting migration list...');
    $list = $this->getMigrationList($this->input()->getOption('tag'));
    if (empty($list) || empty($list = reset($list))) {
      $this->logger()->error(sprintf(
        "No migrations found with the given tag %s.",
        $this->input()->getOption('tag')
      ));
      return self::EXIT_FAILURE;
    }
    $run_at_the_end = $options['run-at-the-end']
      ? StringUtils::csvToArray($options['run-at-the-end'])
      : array_keys(
        array_filter(
          $list,
          function (MigrationInterface $migration) {
            return in_array(
              $migration->getDestinationConfiguration()['plugin'],
              ['entity:path_alias', 'entity:menu_link_content'],
              TRUE
            );
          }
        )
      );
    $infosection->overWrite('Initializing multi-thread processing...');
    $this->initializeMultiThreadProcessing($list, $run_at_the_end);
    $this->preCheckMigrationDependencies($infosection);

    // Determine how many threads we can use.
    $threads_input = $this->input()->getOption('threads');
    $threads = is_numeric($threads_input) ? (int) $threads_input : 4;
    $this->logger()->notice(sprintf(
      "Starting migration in %s threads",
      $threads
    ));

    // Register our SIGINT and SIGTERM signals. If the main process gets one of
    // these, we have to stop all the subprocesses too.
    if (extension_loaded('pcntl')) {
      pcntl_signal(SIGTERM, static::stopSignalClosure());
      pcntl_signal(SIGINT, static::stopSignalClosure());
      pcntl_async_signals(TRUE);
    }

    // Initialize UI.
    $job_index = 1;
    $this->initProgressBars($threads);
    $this->loggerSection = $this->output()->section();
    $this->setLoggerOutput($this->loggerSection);

    while (!(empty($this->activeProcesses) && empty($this->migrationsToProcess))) {
      try {
        while (count($this->activeProcesses) < $threads && !empty($this->migrationsToProcess)) {
          if (!($migration_id = $this->getNextUnblockedMigration())) {
            if (empty($this->activeProcesses)) {
              // There is no unblocked migration, and we don't have any active
              // processes. This means we cannot continue.
              break 2;
            }
            // Let's wait until a process finishes.
            // @todo Move SmartMigrateHelperTrait::getNextUnblockedMigration's
            //   logging here.
            break 1;
          }
          $this->scheduleMigrationProcess($migration_id, $job_index++);
        }

        $mtime = microtime(TRUE);
        $this->updateProgressBars();
        $this->startIdleProcesses();
        $time_left = ($mtime + self::MT_POLL_WAIT_SECONDS) - microtime(TRUE);
        if ($time_left > 0) {
          usleep((int) ($time_left * 1000 * 1000));
        }
        $this->purgeFinishedProcesses();
      }
      catch (\Exception $e) {
        if (static::isStopSignal($e)) {
          // Do not report temporary error messages.
          $this->errorOutputs = array_intersect_key($this->errorOutputs, $this->migrationsProcessed);

          // Kill subprocesses.
          foreach ($this->activeProcesses as $process_data) {
            if (
              $process_data['process'] instanceof SiteProcess &&
              $process_data['process']->isRunning()
            ) {
              try {
                $process_data['process']->signal(SIGKILL);
              }
              catch (\Throwable $t) {
              }
            }
          }
          break 1;
        }
        throw $e;
      }
    }

    $this->clearProgressBars();
    $this->finalReport();

    if (!empty($this->activeProcesses)) {
      $this->logger()->warning('Migration process has been stopped.');
      return self::EXIT_FAILURE;
    }

    if (!empty($this->migrationsToProcess)) {
      $this->logger()->error('Cannot properly finish import process because the following migrations have unmet (required) dependencies: ' . implode(', ', $this->migrationsToProcess));
      return self::EXIT_FAILURE;
    }

    $this->loggerSection->clear();
    $this->logger()->success('Migration process has been finished.');
    $this->release();
    return self::EXIT_SUCCESS;
  }

  /**
   * Executes one migration.
   *
   * @command smart-migrate:single-import
   *
   * @usage smart-migrate:single-import foo
   *   Executes "foo" migration.
   *
   * @validate-module-enabled migrate
   *
   * @hidden
   */
  public function singleImport(): int {
    $migration_id = getenv(self::SMC_MIGRATION_PLUGIN_ID_ENV);
    $migrations = $migration_id ? $this->migrationPluginManager->createInstances([$migration_id]) : [];
    if (count($migrations) !== 1) {
      $this->logger()->error('Too many, or no migrations found.');
      $return = self::EXIT_FAILURE;
    }
    $migration = reset($migrations);
    $message_to_parent = ['id' => $migration_id];
    $verbosity = $this->output()->getVerbosity();
    if (
      !isset($return) &&
      ($requirement_message = $this->getMigrationRequirementsMessages($migration))
    ) {
      $this->logger()->error($requirement_message);
      $return = static::EXIT_REQUIREMENTS_NOT_MET;
    }

    if (empty($return)) {
      // Switch to user 1. This is what we usually do on the UI as well.
      // The user ID mostly affects the user reference (base) fields of migrated
      // entities which don't have source data - e.g. the owner of user or
      // taxonomy term translations.
      $this->accountSwitcher->switchTo(new UserSession(['uid' => 1]));
      try {
        $follow_up_ids = $this->executeMigration($migration);
        if (!empty($follow_up_ids)) {
          $message_to_parent['follow_ups'] = $follow_up_ids;
        }
        $return = self::EXIT_SUCCESS;
      }
      catch (\Exception $e) {
        $this->logger()->critical($e->getMessage());
        $return = self::EXIT_PROCESS_CRITICAL;
      }
      $this->accountSwitcher->switchBack();
    }

    $this->output()->setVerbosity(OutputInterface::VERBOSITY_NORMAL);
    $this->output()->writeln(json_encode($message_to_parent));
    $this->output()->setVerbosity($verbosity);

    return $return;
  }

  /**
   * Returns the sorted and filtered list of migrations with the given tag.
   *
   * @param string $tag
   *   The migration tag of the migrations.
   *
   * @return \Drupal\migrate\Plugin\MigrationInterface[][]
   *   List of the discovered and executable migrations.
   */
  protected function getMigrationList(string $tag): array {
    $grouped_list = $this->doGetGroupedMigrationList(NULL, $tag);
    $grouped_list = $this->filterMigrationList($grouped_list);
    return $this->orderMigrationList($grouped_list, TRUE);
  }

  /**
   * Changes the logger output to the specified output.
   *
   * @param \Symfony\Component\Console\Output\OutputInterface|null $output
   *   The output to use by the logger.
   */
  private function setLoggerOutput(OutputInterface $output = NULL): void {
    $this->logger()->setErrorStream($output ?? $this->output());
    $this->logger()->setOutputStream($output ?? $this->output());
  }

  /**
   * Adds migrate source data to settings.php.
   *
   * @command smart-migrate:configure
   *
   * @option db-url
   *   Url of database connection. Format:
   *   "driver://username:password@host:port/$database#table_prefix".
   * @option source-base-path
   *   Absolute path to the source Drupal instance's web root. Necessary for
   *   migrating files. The path of the files will be calculated with also
   *   evaluating the files_public_path variable of the source Drupal instance,
   *   so please do not add "sites/default/files" to the end of this path.
   * @option source-files-private
   *   Absolute path to the source Drupal instance's private file directory.
   * @option sites-subdir
   *   Name of directory under <info>sites</info> which should be configured.
   *   Optional, defaults to the 'default' directory.
   * @option connection
   *   Name of the database connection to use. Defaults to <info>migrate</info>.
   *
   * @usage smart-migrate:configure --db-url=mysql://user:password@127.0.0.1/foo
   *   Adds the 'foo' mysql database connection as migration source to the
   *   settings.php.
   * @usage smart-migrate:configure --source-base-path=/web/drupal-7
   *   Sets the '/web/drupal-7' directory as migration source webroot.
   * @usage smart-migrate:configure --source-files-private=/web/private
   *   Sets the '/web/private' directory as migration source path of private
   *   files.
   * @usage smart-migrate:configure --db-url=sqlite/localhost/test.sqlite --source-base-path=/web/drupal-7 --source-files-private=/web/private
   *   Multiple options could set in one command.
   *
   * @aliases smconf,sm-conf
   */
  public function migrationConfigure(array $options = [
    'db-url' => self::REQ,
    'source-base-path' => self::REQ,
    'source-files-private' => self::REQ,
    'sites-subdir' => self::OPT,
    'connection' => 'migrate',
  ]): void {
    $db_url = $this->input()->getOption('db-url');
    $files_source = $this->input()->getOption('source-base-path');
    $files_private_source = $this->input()->getOption('source-files-private');

    // If all options are empty, throw an exception.
    if (!$db_url && !$files_source && !$files_private_source) {
      throw new \LogicException('To configure any value for migration, one of --db-url, --source-base-path or --source-files-private options must be defined.');
    }

    // Handle aliases, if any.
    // @see SiteInstallCommands::pre().
    $alias_record = $this->siteAliasManager()->getSelf();
    $root = $alias_record->root();
    if (!($dir = $this->input()->getOption('sites-subdir'))) {
      $dir = $this->getSitesSubdirFromUri($root, $alias_record->get('uri'));
    }

    if (!$dir) {
      throw new \Exception('Could not determine target sites directory for site to install. Use --sites-subdir to specify.');
    }

    $sites_subdir = Path::join('sites', $dir);
    $settingsfile = Path::join($sites_subdir, 'settings.php');
    $settingsfile_write = Path::join($root, $settingsfile);

    if ($db_url) {
      $connection = $this->input()->getOption('connection');
      if (
        Database::getConnectionInfo($connection) &&
        $db_url !== $this->getDatabaseUrl(Database::getConnection('default', $connection)) &&
        !$this->io()->confirm(dt('The given database connection already exists. Do you really want to override?'))
      ) {
        throw new UserAbortException();
      }
      $connection_info = Database::convertDbUrlToConnectionInfo($db_url, $root);
      $settings['databases'][$connection]['default'] = (object) [
        'value' => $connection_info,
        'required' => TRUE,
      ];
    }

    if ($files_source) {
      $current_base_path = Settings::get('source_base_path');
      if (
        $current_base_path &&
        $current_base_path !== $files_source &&
        !$this->io()->confirm(dt(
          "The source base path is already configured, and it is set to '!value'. Do you really want to override?",
          ['!value' => $current_base_path]
        ))
      ) {
        throw new UserAbortException();
      }
      $settings['settings']['source_base_path'] = (object) [
        'value' => $files_source,
        'required' => TRUE,
      ];
    }

    if ($files_private_source) {
      $current_private_source = Settings::get('source_private_file_path');
      if (
        $current_private_source &&
        $current_private_source !== $files_private_source &&
        !$this->io()->confirm(dt(
          "The source private files path is already configured, and it is set to '!value'. Do you really want to override?",
          ['!value' => $current_private_source]
        ))
      ) {
        throw new UserAbortException();
      }
      $settings['settings']['source_private_file_path'] = (object) [
        'value' => $files_private_source,
        'required' => TRUE,
      ];
    }

    if (!is_writable($settingsfile_write) && !chmod($settingsfile_write, 0644)) {
      throw new \Exception(sprintf(
        "Cannot make '%s' file writable. Please modify file permissions manually and re-execute the command!",
        $settingsfile_write
      ));
    }

    include_once $root . '/core/includes/install.inc';
    drupal_rewrite_settings($settings, $settingsfile_write);

    if ($db_url) {
      $this->state->set('smart_migrate_cli.source_connection', [
        'key' => $connection,
        'target' => 'default',
      ]);
      $this->state->set('migrate.fallback_state_key', 'smart_migrate_cli.source_connection');
    }

    $this->logger()->success('Migration configuration was written to settings.php.');
  }

  /**
   * Determine an appropriate site subdir name to use for the provided URI.
   *
   * @param string $root
   *   The Drupal root.
   * @param string $uri
   *   The site URI.
   *
   * @return string|null
   *   The name of the directory of the specified Drupal site, or NULL if it
   *   cannot be determined.
   */
  protected function getSitesSubdirFromUri(string $root, string $uri): ?string {
    $dir = strtolower($uri);
    // Always accept simple uris (e.g. 'dev', 'stage', etc).
    if (preg_match('#^[a-z0-9_-]*$#', $dir)) {
      return $dir;
    }
    // Strip off the protocol from the provided uri -- however,
    // now we will require that the sites subdir already exist.
    $dir = preg_replace('#[^/]*/*#', '', $dir);
    if ($dir && file_exists(Path::join($root, $dir))) {
      return $dir;
    }
    // Find the dir from sites.php file.
    $sites_file = $root . '/sites/sites.php';
    if (file_exists($sites_file)) {
      $sites = [];
      include $sites_file;
      if (!empty($sites) && array_key_exists($uri, $sites)) {
        return $sites[$uri];
      }
    }
    // Fall back to default directory if it exists.
    if (file_exists(Path::join($root, 'sites', 'default'))) {
      return 'default';
    }
    return NULL;
  }

  /**
   * Returns the database URL of the given connection.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The connection.
   *
   * @return string
   *   The database URL representation of the given connection.
   */
  protected function getDatabaseUrl(Connection $connection): string {
    $connection_info = $connection->getConnectionOptions();
    $driver = $connection_info['driver'];
    $user = $connection_info['username'] ?? NULL;
    $pass = $connection_info['password'] ?? NULL;
    $prefix = $connection_info['prefix'];
    $host = strpos($connection_info['database'], DRUPAL_ROOT) === 0
      ? 'localhost'
      : $connection_info['host'] ?? NULL;
    $port = $connection_info['port'] ?? NULL;
    $db_name = strpos($connection_info['database'], DRUPAL_ROOT) === 0
      ? substr($connection_info['database'], strlen(DRUPAL_ROOT) + 1)
      : $connection_info['database'];
    $server = implode(
      '@',
      array_filter([
        implode(':', array_filter([$user, $pass])),
        implode(':', array_filter([$host, $port])),
      ])
    );
    $table_and_prefix = implode('#', array_filter([
      $db_name,
      $prefix,
    ]));
    return "$driver://$server/$table_and_prefix";
  }

  /**
   * Closure called when operation is aborted.
   *
   * @return \Closure
   *   Closure called when operation is aborted.
   */
  protected static function stopSignalClosure(): \Closure {
    return function () {
      throw new \Exception(self::MT_ID);
    };
  }

  /**
   * Determines whether the throwable is a stop signal.
   *
   * @return bool
   *   Whether the throwable is a stop signal.
   */
  protected static function isStopSignal(\Throwable $t): bool {
    return $t->getMessage() === self::MT_ID;
  }

}
