<?php

declare(strict_types = 1);

namespace Drupal\smart_migrate_cli\Traits;

use Consolidation\SiteProcess\SiteProcess;
use Drupal\migrate\MigrateMessageInterface;
use Drupal\migrate\Plugin\migrate\id_map\Sql;
use Drupal\migrate\Plugin\MigrationInterface;
use Drupal\migrate_drupal\Plugin\MigrationWithFollowUpInterface;
use Drupal\smart_migrate_cli\SmartMigrateExecutable;
use Drush\Drupal\Migrate\MigrateMessage;
use Drush\Drush;
use Symfony\Component\Console\Helper\ProgressBar;
use Symfony\Component\Console\Helper\Table;
use Symfony\Component\Console\Helper\TableCell;
use Symfony\Component\Console\Helper\TableSeparator;
use Symfony\Component\Console\Output\ConsoleSectionOutput;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Terminal;

/**
 * Basic and complex routines for smart migrate Drush commands.
 */
trait SmartMigrateHelperTrait {

  use MigrateRunnerTrait;

  /**
   * The last "wait" message logged to output.
   *
   * @var string
   */
  private $lastWaitMessage;

  /**
   * Stores the IDs of migrations which are blocking scheduler.
   *
   * @var string[]
   */
  private $schedulerBlockedBy;

  /**
   * IDs of migrations which should be executed at the end of the process.
   *
   * @var string[]
   */
  private $migrationsToRunAtTheEnd;

  /**
   * Requirements of messages.
   *
   * @var array[]
   */
  private $migrationsRequirements;

  /**
   * Returns a Drush site process which executes a single migration.
   *
   * @param string $migration_id
   *   The plugin ID of the migration to execute.
   *
   * @return \Consolidation\SiteProcess\SiteProcess
   *   The process of the migration. It should be started or ran manually.
   */
  protected function migrationProcess(string $migration_id): SiteProcess {
    return Drush::drush($this->siteAliasManager()->getSelf(), 'smart-migrate:single-import')
      ->setEnv(getenv() + [
        static::SMC_MIGRATION_PLUGIN_ID_ENV => $migration_id,
      ])
      ->enableOutput();
  }

  /**
   * Determines whether the migration is runnable.
   *
   * @param string $plugin_id
   *   The ID of the migration.
   * @param bool $include_optionals
   *   Whether optional dependencies also should be evaluated.
   *
   * @return bool
   *   Whether the migration is runnable or not.
   */
  protected function migrationIsRunnable(string $plugin_id, bool $include_optionals): bool {
    $temp = clone $this->list[$plugin_id];
    $deps = $temp->getMigrationDependencies();
    $requirements = $temp->getRequirements();
    $requirements_key = $include_optionals
      ? 'with_optionals'
      : 'required_only';
    $missings_from_previous_check = $this->migrationsRequirements[$plugin_id][$requirements_key] ?? NULL;
    // We already did a full check, so we know what was missing. Let's have an
    // early return!
    if (is_array($missings_from_previous_check)) {
      unset($temp);
      $unmet = array_diff($requirements, $this->migrationsProcessed);
      return empty($unmet);
    }

    $optionals = array_diff($deps['optional'], [$plugin_id]);
    if (
      ($optionals_present = array_intersect($optionals, array_keys($this->list))) &&
      $include_optionals
    ) {
      $requirements = array_unique(
        array_merge(
          $requirements,
          array_combine($optionals_present, $optionals_present)
        ),
      );
    }

    if (!isset($return)) {
      $temp = $this->list[$plugin_id];
      $temp->set('requirements', $requirements);
      if (!empty($missings = $this->migrationPluginManager->getMissingRequirements($temp))) {
        $return = FALSE;
      }
    }

    if (isset($missings)) {
      $this->migrationsRequirements[$plugin_id][$requirements_key] = $missings;
      if (!$include_optionals && empty($optionals_present)) {
        $this->migrationsRequirements[$plugin_id]['with_optionals'] = $missings;
      }
    }
    unset($temp);
    $return = $return ?? TRUE;
    return $return ?? TRUE;
  }

  /**
   * Returns the ID of the next unblocked migration, or NULL.
   *
   * @return string|null
   *   The ID of the next unblocked migration (which all requirements are met),
   *   or NULL if there are no migrations scheduled for executing, or none of
   *   the remaining migration's requirements are met.
   */
  protected function getNextUnblockedMigration(): ?string {
    // Early return for blocking processes.
    static $last_next;
    if (
      !isset($last_next) &&
      !empty($this->schedulerBlockedBy) &&
      $this->schedulerBlockedBy === array_keys($this->activeProcesses)
    ) {
      // No change.
      return NULL;
    }
    $next_unblocked_plugin_id = NULL;
    $logger_message = NULL;

    foreach ($this->migrationsToProcess as $plugin_id) {
      if (!$this->list[$plugin_id] instanceof MigrationInterface) {
        $this->list[$plugin_id] = $this->getFreshMigrationInstance($plugin_id);
        $this->initializeMapTables($this->list[$plugin_id]);
      }

      // Some migrations should only be returned if nothing else left.
      if (
        !in_array($plugin_id, $this->migrationsToRunAtTheEnd, TRUE) &&
        $this->migrationIsRunnable($plugin_id, TRUE)
      ) {
        $next_unblocked_plugin_id = $plugin_id;
        break 1;
      }
    }

    // Maybe every other migration depends on the run-as-ast ones?
    if (!$next_unblocked_plugin_id && !$this->activeProcesses) {
      foreach (array_intersect_key($this->migrationsToProcess, $this->migrationsToRunAtTheEnd) as $plugin_id) {
        if ($this->migrationIsRunnable($plugin_id, TRUE)) {
          // Yes, it seems.
          $this->migrationsToRunAtTheEnd = [];
          $next_unblocked_plugin_id = $plugin_id;
          break 1;
        }
      }
    }

    // No migration left which required and optional dependencies are met.
    // Let's check only required dependencies, but only if active processes is
    // empty.
    if (!$next_unblocked_plugin_id) {
      if (!$this->activeProcesses) {
        foreach ($this->migrationsToProcess as $plugin_id) {
          if ($this->migrationIsRunnable($plugin_id, FALSE)) {
            $next_unblocked_plugin_id = $plugin_id;
            break 1;
          }
        }
      }
      else {
        $migration_ids = array_reduce(
          $this->activeProcesses,
          function (array $carry, array $item) {
            $carry[] = $item['mid'];
            return $carry;
          },
          []
        );
        $message_template = count($this->activeProcesses) > 1
          ? "Waiting for one of processes %s be finished..."
          : "Waiting for %s be finished...";
        $logger_message = sprintf(
          $message_template,
          implode(', ', $migration_ids)
        );
        if ($this->lastWaitMessage !== $logger_message) {
          $this->schedulerBlockedBy = array_keys($this->activeProcesses);
        }
      }
    }

    if ($next_unblocked_plugin_id) {
      $this->clearLoggerSection();
    }

    if ($logger_message && $this->lastWaitMessage !== $logger_message) {
      $this->lastWaitMessage = $logger_message;
      $this->clearLoggerSection();
      $this->logger()->notice($logger_message);
    }

    $last_next = $next_unblocked_plugin_id;
    return $next_unblocked_plugin_id;
  }

  /**
   * Clears logger section if it is a "thing".
   */
  protected function clearLoggerSection(): void {
    if (isset($this->loggerSection) && $this->loggerSection instanceof OutputInterface) {
      $this->loggerSection->clear();
    }
  }

  /**
   * Initializes the required number of progress bars.
   *
   * Also adds empty lines above and below the bars to increase UX.
   *
   * @param int $threads
   *   Number of the threads. Each thread will have a progress bar.
   */
  private function initProgressBars(int $threads): void {
    $this->barPlaceholderSections['before'] = $this->output()->section();
    $this->barPlaceholderSections['before']->writeln('');
    for ($i = 1; $i <= $threads; $i++) {
      $section = $this->output()->section();
      $section->write(self::EMPTY_PROGRESS_BAR_TEXT);
      $bar = new ProgressBar($section, 0, 0.5);
      $bar->setFormat(static::MT_PROGRESS_BAR_NOMAX);
      $bar->setRedrawFrequency(1);
      $bar->maxSecondsBetweenRedraws(0.8);
      $bar->setBarWidth(static::PROGRESS_BAR_WIDTH);
      $this->progressBars[$i] = [
        'bar' => $bar,
        'section' => $section,
      ];
    }
    $this->barPlaceholderSections['after'] = $this->output()->section();
    $this->barPlaceholderSections['after']->writeln('');
  }

  /**
   * Stops every progress bar and clears their sections.
   */
  private function clearProgressBars(): void {
    foreach ($this->progressBars as $data) {
      if ($data['bar'] instanceof ProgressBar) {
        $data['bar']->finish();
      }
      $data['section']->clear();
    }

    foreach ($this->barPlaceholderSections as $section) {
      $section->clear();
    }
  }

  /**
   * Schedules a subprocess migrating the plugin with the specified ID.
   *
   * @param string $migration_id
   *   The plugin ID of the migration to execute.
   * @param int $job_index
   *   The index of the job.
   */
  private function scheduleMigrationProcess(string $migration_id, int $job_index): void {
    unset($this->migrationsToProcess[$migration_id]);
    $this->startProgressBar($migration_id);
    $this->activeProcesses[$job_index] = [
      'mid' => $migration_id,
      'process' => $this->migrationProcess($migration_id),
    ];
  }

  /**
   * Starts every idle subprocesses.
   */
  private function startIdleProcesses(): void {
    foreach ($this->activeProcesses as $process_data) {
      assert($process_data['process'] instanceof SiteProcess);
      if (!$process_data['process']->isRunning()) {
        $process_data['process']->start();
      }
    }
  }

  /**
   * Starts a progress bar for the migration with the given plugin ID.
   *
   * @param string $migration_id
   *   The plugin ID of the migration.
   */
  private function startProgressBar(string $migration_id): void {
    $source_count = $processed_count = 0;
    if ($this->list[$migration_id] instanceof MigrationInterface) {
      // Source count can be -1.
      $source_count = max(0, $this->list[$migration_id]->getSourcePlugin()->count());
      $processed_count = $this->list[$migration_id]->getIdMap()->processedCount();
    }

    $free_progress_bars = array_filter(
      $this->progressBars,
      function (array $data) {
        return empty($data['mid']);
      }
    );
    reset($free_progress_bars);
    $free_key = key($free_progress_bars);

    $format = $source_count ? static::MT_PROGRESS_BAR : static::MT_PROGRESS_BAR_NOMAX;

    $this->progressBars[$free_key]['mid'] = $migration_id;
    assert($this->progressBars[$free_key]['bar'] instanceof ProgressBar);
    $this->progressBars[$free_key]['section']->clear();
    $this->progressBars[$free_key]['bar']->setMessage("<info>$migration_id</info>");
    $this->progressBars[$free_key]['bar']->setFormat($format);
    $this->progressBars[$free_key]['bar']->start($source_count);
    $this->progressBars[$free_key]['bar']->setProgress($processed_count);
  }

  /**
   * Stops (finishes) the progress bar of the given migration.
   *
   * @param string $migration_id
   *   The plugin ID of the migration.
   */
  private function stopProgressBar(string $migration_id): void {
    if (is_null($pb_key = $this->getProgressBarKey($migration_id))) {
      return;
    }

    $this->progressBars[$pb_key]['mid'] = NULL;
    assert($this->progressBars[$pb_key]['bar'] instanceof ProgressBar);
    $this->progressBars[$pb_key]['bar']->finish();
    $this->progressBars[$pb_key]['bar']->clear();
    $this->progressBars[$pb_key]['section']->overwrite(self::EMPTY_PROGRESS_BAR_TEXT);
  }

  /**
   * Removes all finished processes from the active process storage.
   *
   * This method also processes and saves migration messages logged during the
   * subrocess and saves them to the main process' error output storage.
   */
  private function purgeFinishedProcesses(): void {
    $realActiveProcesses = array_filter(
      $this->activeProcesses,
      function (array $process_data) {
        $subprocess = $process_data['process'];
        if (!($is_running = $subprocess->isRunning())) {
          $migration_id = $process_data['mid'];
          $process_output = array_filter(explode("\n", $subprocess->getOutput()));
          $last = array_pop($process_output);
          $message = $last ? json_decode($last, TRUE) : [];
          $this->migrationsProcessed[$migration_id] = $migration_id;
          $errors = array_reduce(
            array_filter(explode("\n", $subprocess->getErrorOutput())),
            function (array $carry, string $error) {
              $key = md5($error);
              $carry[$key] = $carry[$key] ?? [
                'count' => 0,
                'message' => $error,
              ];
              $carry[$key]['count'] += 1;
              return $carry;
            },
            []
          );
          if (empty($errors)) {
            unset($this->errorOutputs[$migration_id]);
          }
          else {
            $this->errorOutputs[$migration_id] = $errors;
          }
          if (!empty($message['follow_ups'])) {
            $follow_ups = array_combine($message['follow_ups'], $message['follow_ups']);
            $diff = array_diff_key($follow_ups, $this->list);
            if ($diff) {
              $this->list = array_merge(
                $this->list,
                $diff
              );
              $this->migrationsToProcess += array_combine(array_keys($diff), array_keys($diff));
            }
          }

          switch ($subprocess->getExitCode()) {
            // We already handle these exit codes.
            case self::EXIT_SUCCESS:
              break;

            case self::EXIT_PROCESS_CRITICAL:
              $this->clearLoggerSection();
              $this->logger()->critical(sprintf(
                "Migration process of '%s' migration threw an exception.",
                $migration_id
              ));
              call_user_func(static::stopSignalClosure());
              break;

            default:
              $this->clearLoggerSection();
              $this->logger()->critical(sprintf(
                "An unexpected error happened during executing '%s' migration",
                $migration_id
              ));
              call_user_func(static::stopSignalClosure());
          }
        }
        return $is_running;
      }
    );

    // Clean up finished processes.
    foreach (array_diff_key($this->activeProcesses, $realActiveProcesses) as $item) {
      $this->stopProgressBar($item['mid']);
    }
    $this->activeProcesses = $realActiveProcesses;
  }

  /**
   * Returns the migration's progress bar's key in progress bar storage.
   *
   * @param string $migration_id
   *   The plugin ID of the migration.
   *
   * @return int|null
   *   The key of the progress bar, or NULL if the given migration does not have
   *   any progress bars.
   */
  private function getProgressBarKey(string $migration_id): ?int {
    $progress_bars = array_filter(
      $this->progressBars,
      function (array $progress_bar_data) use ($migration_id) {
        return ($progress_bar_data['mid'] ?? NULL) === $migration_id;
      }
    );
    reset($progress_bars);
    return $progress_bars ? key($progress_bars) : NULL;
  }

  /**
   * Updates progress bars of the active processes.
   */
  private function updateProgressBars(): void {
    foreach ($this->activeProcesses as $data) {
      $pb_key = $this->getProgressBarKey($data['mid']);
      $processed_count = $this->list[$data['mid']]->getIdMap()->processedCount();
      $this->progressBars[$pb_key]['bar']->setProgress($processed_count);
    }
  }

  /**
   * Renders a final report about the errors recorded during the migration.
   */
  private function finalReport(): void {
    if (empty($errors_per_plugin = array_filter($this->errorOutputs))) {
      return;
    }

    // Always render report.
    if (($verbosity_original = $this->output()->getVerbosity()) < OutputInterface::VERBOSITY_NORMAL) {
      $this->output()->setVerbosity(OutputInterface::VERBOSITY_NORMAL);
    }

    // Get terminal width. This ensures we always use the entire width.
    $terminal = new Terminal();
    $width = $terminal->getWidth();
    $headers = [dt('Error'), dt('Count')];
    $available_width = $width - ((count($headers) * 3) + 1);

    $table = new Table($this->output());
    $table->setHeaders($headers)
      // Error column always use the rest of the space.
      ->setColumnMaxWidth(0, $available_width - 6)
      // Width of the "Count" column is fixed.
      ->setColumnMaxWidth(1, 6)
      ->setColumnWidths([0 => $available_width - 6, 1 => 6]);
    ksort($errors_per_plugin);
    $last = array_key_last($errors_per_plugin);
    foreach ($errors_per_plugin as $migration_id => $errors) {
      $table->addRow(
        [new TableCell("<comment>$migration_id</comment>", ['colspan' => 2])]
      );
      $table->addRow(new TableSeparator());
      foreach ($errors as $error) {
        $table->addRow([
          str_replace(DRUPAL_ROOT, '<DRUPAL_ROOT>', $error['message']),
          $error['count'] ?? 1,
        ]);
      }
      if ($last !== $migration_id) {
        $table->addRow(new TableSeparator());
      }
    }

    $this->logger()->warning('Some migrations triggered errors during processing. Please check them!');
    $table->render();
    $this->output()->writeln('');
    $this->output()->setVerbosity($verbosity_original);
  }

  /**
   * Executes the given migration plugin instance.
   *
   * Simplified version of Drush's version in MigrateRunnerCommands.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance to execute.
   *
   * @return string[]
   *   IDs of follow up migration, if any.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *
   * @see \Drush\Drupal\Commands\core\MigrateRunnerCommands::executeMigration()
   */
  protected function executeMigration(MigrationInterface $migration): array {
    $this->prepareMigration($migration);

    $executable = new SmartMigrateExecutable($migration, $this->getMigrateMessage(), $this->output());
    drush_op([$executable, 'import']);
    if ($failure_count = $migration->getIdMap()->errorCount()) {
      $this->logger()->warning(dt(
        '!name migration: ID map plugin reports !count failed rows.',
        ['!name' => $migration->id(), '!count' => $failure_count]
      ));
    }

    // After the migration on which they depend has been successfully executed,
    // the follow-up migrations should be immediately added to the list of the
    // migrations to be executed.
    $follow_ups = $migration instanceof MigrationWithFollowUpInterface
      ? $migration->generateFollowUpMigrations()
      : [];

    return array_keys($follow_ups);
  }

  /**
   * Checks whether the given migration plugin's requirements are met.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration to check.
   *
   * @return bool
   *   Whether the given migration plugin's requirements are met.
   */
  protected function migrationRequirementsAreMet(MigrationInterface $migration): bool {
    return $this->migrationPluginManager->migrationRequirementsMet($migration, TRUE);
  }

  /**
   * Gets the unmet requirements problem message of the given migration, if any.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration to check.
   *
   * @return string|null
   *   The requirements issues of the plugin as human readable string, or NULL
   *   if the plugin's requirements are met.
   */
  protected function getMigrationRequirementsMessages(MigrationInterface $migration): ?string {
    return $this->migrationPluginManager->getMigrationRequirementsMessages($migration);
  }

  /**
   * Initializes the multi thread migration process.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface[] $list
   *   List of the migrations to execute.
   * @param string[] $run_at_the_end
   *   IDs of migrations to execute at the end of the process.
   */
  protected function initializeMultiThreadProcessing(array $list, array $run_at_the_end = []): void {
    $this->list = $list;

    // Force-create migrate map tables.
    array_walk($this->list, [$this, 'initializeMapTables']);

    // Unstuck migrations.
    array_walk($this->list, [$this, 'unstuckMigration']);

    $already_processed = array_filter(
      $this->list,
      [$this, 'migrationIsFullyProcessed']
    );

    $this->migrationsProcessed = array_combine(
      array_keys($already_processed),
      array_keys($already_processed)
    );

    $needs_process = array_diff(array_keys($this->list), $this->migrationsProcessed);

    // Populate our local storage which stores the list of the migrations which
    // should be executed.
    $this->migrationsToProcess = array_combine($needs_process, $needs_process);

    // Some core migrations have the best results if they are executed at the
    // end of the process.
    $this->migrationsToRunAtTheEnd = array_combine($run_at_the_end, $run_at_the_end);
  }

  /**
   * Initialized the migration map tables of the given migration.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance.
   */
  protected function initializeMapTables(MigrationInterface $migration): void {
    $clone = clone $migration;
    if (($id_map = $clone->getIdMap()) instanceof Sql) {
      $id_map->getDatabase();
    }
    unset($clone);
  }

  /**
   * Unstucks the given migration if it is stuck.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance.
   */
  protected function unstuckMigration(MigrationInterface $migration): void {
    if ($migration->getStatus() === MigrationInterface::STATUS_IMPORTING) {
      $migration->setStatus(MigrationInterface::STATUS_IDLE);
    }
  }

  /**
   * Determines whether the rows of the migration are fully processed or not.
   *
   * @param \Drupal\migrate\Plugin\MigrationInterface $migration
   *   The migration plugin instance.
   *
   * @return bool
   *   Whether the rows of the migration are fully processed or not. Fully
   *   processed row means that it was saved and does not need update, or it
   *   was skipped on purpose.
   */
  protected function migrationIsFullyProcessed(MigrationInterface $migration): bool {
    $clone = clone $migration;
    try {
      $source_count = $clone->getSourcePlugin()->count();
      // If the source is uncountable, we have no way of knowing if it's
      // complete, so stipulate that it is.
      $fully_processed_count = 0;
      if ($source_count > 0) {
        $id_map = $clone->getIdMap();
        $fully_processed_count = $id_map->processedCount() - ($id_map->updateCount() + $id_map->errorCount());
      }

      $return = $source_count <= $fully_processed_count;
    }
    catch (\Exception $e) {
      unset($this->list[$migration->id()]);
      $return = FALSE;
    }
    unset($clone);
    return $return;
  }

  /**
   * Returns the migrate message logger.
   *
   * @return \Drupal\migrate\MigrateMessageInterface
   *   The migrate message logger.
   */
  protected function getMigrateMessage(): MigrateMessageInterface {
    if (!isset($this->migrateMessage)) {
      $this->migrateMessage = new MigrateMessage($this->logger());
    }
    return $this->migrateMessage;
  }

  /**
   * Checks migration's dependencies and prefills migrationRequirements.
   */
  protected function preCheckMigrationDependencies(ConsoleSectionOutput $output): void {
    $output->clear();
    $migration_ids_to_check = array_diff(array_keys($this->list), $this->migrationsProcessed);
    if ($migration_ids_to_check < 200) {
      return;
    }
    $progressbar = new ProgressBar($output);
    $progressbar->setBarWidth(static::PROGRESS_BAR_WIDTH);
    $progressbar->setMaxSteps(count($migration_ids_to_check));
    $progressbar->setFormat(static::MT_PROGRESS_BAR);
    $progressbar->setMessage('Preliminary checking migration dependencies...');
    $progressbar->start();
    foreach ($migration_ids_to_check as $plugin_id) {
      $this->migrationIsRunnable($plugin_id, FALSE);
      $this->migrationIsRunnable($plugin_id, TRUE);
      $progressbar->advance();
    }
    $progressbar->finish();
    $output->clear();
  }

}
